import requests
from lxml import etree
from retry import retry


@retry(tries=3, delay=2, backoff=1.2, jitter=(1, 3))
def download_xml_from_ena(ena_url) -> etree.XML:
    """Download and parse XML from ENA"""
    try:  # catches any kind of request error, including non-20X status code
        response = requests.get(ena_url)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        raise e
    root = etree.XML(bytes(response.text, encoding='utf-8'))
    return root


def get_assembly_name_and_taxonomy_id(assembly_accession):
    xml_root = download_xml_from_ena(f'https://www.ebi.ac.uk/ena/browser/api/xml/{assembly_accession}')
    xml_assembly = xml_root.xpath('/ASSEMBLY_SET/ASSEMBLY')
    if len(xml_assembly) == 0:
        raise ValueError(f'Assembly {assembly_accession} not found in ENA')
    assembly_name = xml_assembly[0].get('alias')
    taxonomy_id = int(xml_assembly[0].xpath('TAXON/TAXON_ID')[0].text)
    return assembly_name, taxonomy_id


def get_scientific_name_and_common_name(taxonomy_id):
    xml_root = download_xml_from_ena(f'https://www.ebi.ac.uk/ena/browser/api/xml/{taxonomy_id}')
    xml_taxon = xml_root.xpath('/TAXON_SET/taxon')
    if len(xml_taxon) == 0:
        raise ValueError(f'Taxonomy {taxonomy_id} not found in ENA')
    scientific_name = xml_taxon[0].get('scientificName')
    optional_common_name = xml_taxon[0].get('commonName')
    return scientific_name, optional_common_name



#!/usr/bin/env python
# Copyright 2020 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import requests
from functools import cached_property
from ebi_eva_common_pyutils.logger import AppLogger
from retry import retry


class HALNotReadyError(Exception):
    pass


class HALCommunicator(AppLogger):
    """
    This class helps navigate through REST API that uses the HAL standard.
    """
    acceptable_code = [200, 201]

    def __init__(self, auth_url, bsd_url, username, password):
        self.auth_url = auth_url
        self.bsd_url = bsd_url
        self.username = username
        self.password = password

    def _validate_response(self, response):
        """Check that the response has an acceptable code and raise if it does not"""
        if response.status_code not in self.acceptable_code:
            self.error(response.request.method + ': ' + response.request.url + " with " + str(response.request.body))
            self.error("headers: {}".format(response.request.headers))
            self.error("<{}>: {}".format(response.status_code, response.text))
            raise ValueError('The HTTP status code ({}) is not one of the acceptable codes ({})'.format(
                str(response.status_code), str(self.acceptable_code))
            )
        return response

    @cached_property
    def token(self):
        """Retrieve the token from the REST API then cache it for further querying"""
        response = requests.get(self.auth_url, auth=(self.username, self.password))
        self._validate_response(response)
        return response.text

    @retry(exceptions=(ValueError, requests.RequestException), tries=3, delay=2, backoff=1.2, jitter=(1, 3))
    def _req(self, method, url, **kwargs):
        """Private method that sends a request using the specified method. It adds the headers required by bsd"""
        headers = kwargs.pop('headers', {})
        headers.update({'Accept': 'application/hal+json'})
        if self.token is not None:
            headers.update({'Authorization': 'Bearer ' + self.token})
        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
        response = requests.request(
            method=method,
            url=url,
            headers=headers,
            **kwargs
        )
        self._validate_response(response)
        return response

    def follows(self, query, json_obj=None, method='GET', url_template_values=None, join_url=None, **kwargs):
        """
        Finds a link within the json_obj using a query string or list, modify the link using the
        url_template_values dictionary then query the link using the method and any additional keyword argument.
        If the json_obj is not specified then it will use the root query defined by the base url.
        """
        all_pages = kwargs.pop('all_pages', False)

        if json_obj is None:
            json_obj = self.root
        # Drill down into a dict using dot notation
        _json_obj = json_obj
        if isinstance(query, str):
            query_list = query.split('.')
        else:
            query_list = query
        for query_element in query_list:
            if query_element in _json_obj:
                _json_obj = _json_obj[query_element]
            else:
                raise KeyError('{} does not exist in json object'.format(query_element, _json_obj))
        if not isinstance(_json_obj, str):
            raise ValueError('The result of the query_string must be a string to use as a url')
        url = _json_obj
        # replace the template in the url with the value provided
        if url_template_values:
            for k, v in url_template_values.items():
                url = re.sub('{(' + k + ')(:.*)?}', v, url)
        if join_url:
            url += '/' + join_url
        text_only = False
        if 'text_only' in kwargs and kwargs.get('text_only'):
            text_only = kwargs.pop('text_only')
        # Now query the url
        response = self._req(method, url, **kwargs)
        if text_only:
            return response.text

        json_response = response.json()
        # Depaginate the call if requested
        if all_pages is True:
            # This depagination code will iterate over all the pages available until the pages comes back  without a
            # next page. It stores the embedded elements in the initial query's json response
            content = json_response
            while 'next' in content.get('_links'):
                content = self._req(method, content.get('_links').get('next').get('href'), **kwargs).json()
                for key in content.get('_embedded'):
                    json_response['_embedded'][key].extend(content.get('_embedded').get(key))
            # Remove the pagination information as it is not relevant to the depaginated response
            if 'page' in json_response: json_response.pop('page')
            if 'first' in json_response['_links']: json_response['_links'].pop('first')
            if 'last' in json_response['_links']: json_response['_links'].pop('last')
            if 'next' in json_response['_links']: json_response['_links'].pop('next')
        return json_response

    def follows_link(self, key, json_obj=None, method='GET', url_template_values=None, join_url=None, **kwargs):
        """
        Same function as follows but construct the query_string from a single keyword surrounded by '_links' and 'href'.
        """
        return self.follows(('_links', key, 'href'),
                            json_obj=json_obj, method=method, url_template_values=url_template_values,
                            join_url=join_url, **kwargs)

    @cached_property
    def root(self):
        return self._req('GET', self.bsd_url).json()

    @property
    def communicator_attributes(self):
        raise NotImplementedError

class WebinHALCommunicator(HALCommunicator):
    """Class to navigate BioSamples API using Webin authentication."""

    @cached_property
    def token(self):
        """Retrieve the token from the ENA Webin REST API then cache it for further querying"""
        response = requests.post(self.auth_url,
                                 json={"authRealms": ["ENA"], "password": self.password,
                                       "username": self.username})
        self._validate_response(response)
        return response.text

    @property
    def communicator_attributes(self):
        return {'webinSubmissionAccountId': self.username}


class NoAuthHALCommunicator(HALCommunicator):
    """Class to navigate BioSamples API without authentication."""

    def __init__(self, bsd_url):
        super(NoAuthHALCommunicator, self).__init__(None, bsd_url, None, None)

    @cached_property
    def token(self):
        """No auth token, so errors will be raised if auth is required for requests"""
        return None



# Copyright 2020 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import logging.config
import logging.handlers
from sys import stdout, stderr
from cached_property import cached_property


class LoggingConfiguration:
    """
    This class provides an all in one management of all loggers in the stack. By default it pulls existing loggers,
    stores additional ones along with handlers and formatters.
    """

    default_fmt = '[%(asctime)s][%(name)s][%(levelname)s] %(message)s'
    default_datefmt = '%Y-%b-%d %H:%M:%S'

    def __init__(self, use_existing_logger=True, log_level=logging.INFO):
        self.blank_formatter = logging.Formatter()
        self.handlers = set()
        if use_existing_logger:
            # retrieve all third party loggers
            self.loggers = dict((name, logger)
                                for name, logger in logging.root.manager.loggerDict.items()
                                if isinstance(logger, logging.Logger))
        else:
            self.loggers = {}
        self._log_level = log_level

    @cached_property
    def formatter(self):
        return self.default_formatter

    @cached_property
    def default_formatter(self):
        return logging.Formatter(
            fmt=self.default_fmt,
            datefmt=self.default_datefmt
        )

    def get_logger(self, name, level=logging.NOTSET):
        """
        Return a logging.Logger object with formatters and handlers added.
        :param name: Name to assign to the logger (usually __name__)
        :param int level: Log level to assign to the logger upon creation
        """
        if name in self.loggers:
            logger = self.loggers[name]
        else:
            logger = logging.getLogger(name)
            self.loggers[name] = logger

        logger.setLevel(level or self._log_level)
        for h in self.handlers:
            logger.addHandler(h)

        return logger

    def add_handler(self, handler, level=logging.NOTSET):
        """
        Add a created handler, set its format/level if needed and register all loggers to it
        :param logging.Handler handler:
        :param int level: Log level to assign to the created handler
        """
        handler.setLevel(level or self._log_level)
        handler.setFormatter(self.formatter)
        for name in self.loggers:
            self.loggers[name].addHandler(handler)
        self.handlers.add(handler)

    def add_stdout_handler(self, level=None):
        self.add_handler(logging.StreamHandler(stdout), level=level or self._log_level)

    def add_stderr_handler(self, level=None):
        self.add_handler(logging.StreamHandler(stderr), level=level or self._log_level)

    def add_file_handler(self, filename, level=None):
        self.add_handler(logging.FileHandler(filename=filename), level=level or self._log_level)

    def set_log_level(self, level):
        self._log_level = level
        for h in self.handlers:
            h.setLevel(self._log_level)
        for name in self.loggers:
            self.loggers[name].setLevel(self._log_level)

    def set_formatter(self, formatter):
        """
        Set all handlers to use formatter
        :param logging.Formatter formatter:
        """
        self.__dict__['formatter'] = formatter
        for h in self.handlers:
            h.setFormatter(self.formatter)

    def reset(self):
        """Remove all handlers of existing logger"""
        for l in self.loggers.values():
            while l.handlers:
                l.removeHandler(l.handlers[0])

        while self.handlers:
            h = self.handlers.pop()
            del h


# A logging configuration singleton that will be the only source of logger
logging_config = LoggingConfiguration()


class AppLogger:
    """
    Mixin class for logging. An object subclassing this can log using its class name. Contains a
    logging.Logger object and exposes its log methods.
    """
    log_cfg = logging_config

    def log(self, level, msg, *args, **kwargs):
        self._logger.log(level, msg, *args, **kwargs)

    def debug(self, msg, *args):
        self._logger.debug(msg, *args)

    def info(self, msg, *args):
        self._logger.info(msg, *args)

    def warning(self, msg, *args):
        self._logger.warning(msg, *args)

    def error(self, msg, *args):
        self._logger.error(msg, *args)

    def critical(self, msg, *args):
        self._logger.critical(msg, *args)

    @cached_property
    def _logger(self):
        return self.log_cfg.get_logger(self.__class__.__name__)


# Copyright 2020 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pymongo
from urllib.parse import quote_plus

from pymongo import ReadPreference

from ebi_eva_common_pyutils.command_utils import run_command_with_output
from ebi_eva_common_pyutils.common_utils import merge_two_dicts
from ebi_eva_internal_pyutils.config_utils import get_mongo_uri_for_eva_profile, get_mongo_creds_for_profile


class MongoConfig:
    parameters = None

    def __init__(self, **kwargs):
        self.parameters = kwargs
        if "port" not in kwargs:
            self.parameters["port"] = 27017
        if "host" not in kwargs:
            self.parameters["host"] = "localhost"


def get_mongo_connection_handle(profile: str, settings_xml_file: str,
                                read_concern: str = "majority",
                                read_preference: ReadPreference = ReadPreference.PRIMARY,
                                write_concern: str = "majority") -> pymongo.MongoClient:
    mongo_connection_uri = get_mongo_uri_for_eva_profile(profile, settings_xml_file)
    return pymongo.MongoClient(mongo_connection_uri,
                               readConcernLevel=read_concern,
                               read_preference=read_preference,
                               w=write_concern)


def get_primary_mongo_connection_handle(profile: str, settings_xml_file: str,
                                        read_concern: str = "majority",
                                        read_preference: ReadPreference = ReadPreference.PRIMARY,
                                        write_concern: str = "majority") -> pymongo.MongoClient:
    host, username, password = get_mongo_creds_for_profile(profile, settings_xml_file)
    mongo_connection_uri = "mongodb://{0}:{1}@{2}:{3}/{4}".format(username, quote_plus(password), host,
                                                                  27017, "admin")
    return pymongo.MongoClient(mongo_connection_uri,
                               readConcernLevel=read_concern,
                               read_preference=read_preference,
                               w=write_concern)


def copy_db_with_config(mongo_source_config: MongoConfig, mongo_destination_config: MongoConfig, mongodump_args: dict,
                        mongorestore_args: dict):
    copy_db(merge_two_dicts(mongo_source_config.parameters, mongodump_args),
            merge_two_dicts(mongo_destination_config.parameters, mongorestore_args))


def copy_db(mongodump_args: dict, mongorestore_args: dict):
    mongodump_args_str = " ".join(["--{0} {1}".format(key, value) for key, value in mongodump_args.items()])
    mongorestore_args_str = " ".join(["--{0} {1}".format(key, value) for key, value in mongorestore_args.items()])
    run_command_with_output("Running mongodump", "mongodump " + mongodump_args_str, log_error_stream_to_output=True)
    run_command_with_output("Running mongorestore", "mongorestore " + mongorestore_args_str,
                            log_error_stream_to_output=True)



# Copyright 2020 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from ebi_eva_common_pyutils.ncbi_utils import retrieve_species_scientific_name_from_tax_id_ncbi
from ebi_eva_common_pyutils.network_utils import json_request
from ebi_eva_common_pyutils.logger import logging_config as log_cfg


logger = log_cfg.get_logger(__name__)


def get_scientific_name_from_ensembl(taxonomy_id: int) -> str:
    ENSEMBL_REST_API_URL = "https://rest.ensembl.org/taxonomy/id/{0}?content-type=application/json".format(taxonomy_id)
    response = json_request(ENSEMBL_REST_API_URL)
    if "scientific_name" not in response:
        raise Exception("Scientific name could not be found for taxonomy {0} using the Ensembl API URL: {1}"
                        .format(taxonomy_id, ENSEMBL_REST_API_URL))
    return response["scientific_name"]


def normalise_taxon_scientific_name(taxon_name):
    """
    Match Ensembl representation
    See Clostridium sp. SS2/1 represented as clostridium_sp_ss2_1 in
    ftp://ftp.ensemblgenomes.org/pub/bacteria/release-48/fasta/bacteria_25_collection/clostridium_sp_ss2_1/
    """
    return re.sub('[^0-9a-zA-Z]+', '_', taxon_name.lower())


def get_normalized_scientific_name_from_ensembl(taxonomy_id: int) -> str:
    """Get the scientific name for that taxon"""
    return normalise_taxon_scientific_name(get_scientific_name_from_ensembl(taxonomy_id))


def get_scientific_name_from_taxonomy(taxonomy_id: int, api_key: str=None) -> str:
    """
    Search for a species scientific name based on the taxonomy id.
    Will first attempt to retrieve from Ensembl and then NCBI, if not found returns None.
    """
    try:
        species_name = get_scientific_name_from_ensembl(taxonomy_id)
    except Exception:
        logger.warning("Failed to retrieve scientific name in Ensembl for taxonomy id {0}".format(taxonomy_id))
        species_name = None
    if not species_name:
        species_name = retrieve_species_scientific_name_from_tax_id_ncbi(taxonomy_id, api_key=api_key)
    return species_name



# Copyright 2019 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from functools import lru_cache

from ebi_eva_common_pyutils.logger import logging_config
from ebi_eva_common_pyutils.network_utils import json_request
from ebi_eva_common_pyutils.taxonomy.taxonomy import get_normalized_scientific_name_from_ensembl

logger = logging_config.get_logger(__name__)


def get_supported_asm_from_ensembl(tax_id: int) -> str:
    logger.info(f'Query Ensembl for species name using taxonomy {tax_id}')
    scientific_name_api_param = get_normalized_scientific_name_from_ensembl(tax_id)
    ENSEMBL_REST_API_URL = "http://rest.ensembl.org/info/assembly/{0}?content-type=application/json".format(
        scientific_name_api_param)
    response = json_request(ENSEMBL_REST_API_URL)
    assembly_accession_attribute = 'assembly_accession'
    if assembly_accession_attribute in response:
        return str(response.get(assembly_accession_attribute))
    return None


@lru_cache(maxsize=None)
def get_taxonomy_to_assembly_mapping_from_ensembl_rapid_release():
    """
    Returns a dict mapping taxonomy ID to assembly accession, choosing the most recently released,
    lexicographically last, non-alternate haplotype assembly when multiple are present.
    """
    list_data = json_request('https://ftp.ensembl.org/pub/rapid-release/species_metadata.json')
    results = {}
    for asm_data in list_data:
        tax_id = asm_data['taxonomy_id']
        asm_accession = asm_data['assembly_accession']
        strain = asm_data['strain']
        release_date = datetime.strptime(asm_data['release_date'], '%Y-%m-%d')

        # Skip alternate haplotype assemblies
        if strain and strain.lower() == 'alternate haplotype':
            continue
        # Skip non GCA accession
        if not asm_accession.startswith('GCA_'):
            continue

        # If we haven't seen this taxonomy before, just use this assembly
        if tax_id not in results:
            results[tax_id] = (asm_accession, release_date)
            continue

        current_assembly, current_date = results[tax_id]
        # Keep the more recent assembly, or the lexicographically last one if release dates are equal
        if current_date < release_date or (current_date == release_date and asm_accession > current_assembly):
            results[tax_id] = (asm_accession, release_date)

    return {key: val[0] for key, val in results.items()}


def get_supported_asm_from_ensembl_rapid_release(tax_id: int) -> str:
    # TODO: Replace with API call once supported
    rapid_release_data = get_taxonomy_to_assembly_mapping_from_ensembl_rapid_release()
    return rapid_release_data.get(tax_id, None)



# Copyright 2020 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import psycopg2
from ebi_eva_common_pyutils.logger import logging_config as log_cfg

logger = log_cfg.get_logger(__name__)


def get_all_results_for_query(pg_conn, query):
    with get_result_cursor(pg_conn, query) as pg_cursor:
        results = pg_cursor.fetchall()
    return results


def execute_query(pg_conn, query):
    with get_result_cursor(pg_conn, query) as _:
        pg_conn.commit()


def get_result_cursor(pg_conn, query):
    pg_cursor = pg_conn.cursor()
    pg_cursor.execute(query)
    return pg_cursor


def get_pg_connection_handle(dbname, user, host):
    return psycopg2.connect("dbname='{0}' user='{1}' host='{2}'".format(dbname, user, host))


def index_already_exists_on_table(pg_conn, schema_name, table_name, index_columns):
    index_columns_lower_case = list(map(str.lower, index_columns))
    query = """select unnest(column_names) from (
                    select
                        nmsp.nspname as schema_name,
                        t.relname as table_name,
                        i.relname as index_name,
                        array_agg(a.attname) as column_names,
                        count(*) as number_of_columns
                    from
                        pg_class t,
                        pg_class i,
                        pg_index ix,
                        pg_attribute a,
                        pg_namespace nmsp
                    where
                        t.oid = ix.indrelid
                        and i.oid = ix.indexrelid
                        and a.attrelid = t.oid
                        and a.attnum = ANY(ix.indkey)
                        and t.relkind = 'r'
                        and nmsp.oid = t.relnamespace
                        and nmsp.nspname = '{0}'
                        and t.relname = '{1}'
                        and a.attname in ({2})                        
                    group by    
                        schema_name, table_name, index_name
                    order by
                        t.relname,
                        i.relname
                    ) temp 
        where number_of_columns = {3};
    """.format(schema_name, table_name,
               ",".join(["'{0}'".format(col) for col in index_columns_lower_case]), len(index_columns))
    results = [result[0] for result in get_all_results_for_query(pg_conn, query)]
    return sorted(results) == index_columns_lower_case


def create_index_on_table(pg_conn, schema_name, table_name, index_columns):
    if index_already_exists_on_table(pg_conn, schema_name, table_name, index_columns):
        logger.info("Index on {0} column(s) on {1}.{2} already exists. Skipping..."
                    .format(",".join(list(map(str.lower, sorted(index_columns)))), schema_name, table_name))
    else:
        query = "create index on {0}.{1} ({2})".format(schema_name, table_name,
                                                       ",".join(list(map(str.lower, sorted(index_columns))))
                                                       )
        logger.info("Building index with query: " + query)
        execute_query(pg_conn, query)
        pg_conn.commit()


def vacuum_analyze_table(pg_conn, schema_name, table_name, columns=()):
    query = "vacuum analyze {0}.{1}".format(schema_name, table_name)
    if columns:
        query += "({0})".format(",".join(columns))
    isolation_level_pre_analyze = pg_conn.isolation_level
    try:
        # This is needed for vacuum analyze to work since it can't work inside transactions!
        pg_conn.set_isolation_level(0)
        logger.info("Vacuum analyze with query: " + query)
        execute_query(pg_conn, query)
    except Exception as ex:
        logger.error(ex)
    finally:
        pg_conn.set_isolation_level(isolation_level_pre_analyze)



# Copyright 2020 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import requests
from retry import retry


# TODO: Might be a good idea to re-visit this after a production implementation
#  of the contig-alias resolution project is available
@retry(tries=10, delay=5, backoff=1.2, jitter=(1, 3))
def resolve_contig_accession_to_chromosome_name(contig_accession, line_limit=100):
    """
    Given a Genbank contig accession, get the corresponding chromosome name from the ENA Text API
    which returns results in a EMBL Flatfile format

    :param contig_accession: Genbank contig accession (ex:  CM003032.1)
    :param line_limit: number of lines to parse in the EMBL Flatfile result to find the chromosome before giving up
    :return: Chromosome name (ex: 12 when given an accession CM003032.1)
    """
    ENA_TEXT_API_URL = "https://www.ebi.ac.uk/ena/browser/api/text/{0}?lineLimit={1}&annotationOnly=true"
    response = requests.get(ENA_TEXT_API_URL.format(contig_accession, line_limit))
    response_lines = response.content.decode("utf-8").split("\n")
    num_lines = len(response_lines)

    features_section_found, source_line_found = False, False
    chosen_response = []
    line_index = 0
    # Look for the "source" feature under the "Features" section in the text response
    while line_index < num_lines:
        line = response_lines[line_index]
        if not (features_section_found or line.lower().startswith("fh   key")):
            line_index += 1
            continue
        features_section_found = True
        # Based on "Data item positions" described here, http://www.insdc.org/files/feature_table.html#3.4.2
        # the sixth character represents the start of the feature key
        if not (source_line_found or line[5:].lower().startswith("source")):
            line_index += 1
            continue
        source_line_found = True
        if line[21:].startswith("/"):
            assembled_line = line.strip()
            line_index += 1
            # Assemble text spread across multiple lines until
            # we hit the next qualifier (starts with /) or the next section
            while line_index < num_lines and \
                    not (response_lines[line_index][21:].startswith("/")
                         or response_lines[line_index][5:6].strip() != ''):
                line = response_lines[line_index]
                assembled_line += " " + line[21:].strip()
                line_index += 1

            # Fall back to organelle in case of MT/Chloroplast accessions
            # and the reference notes in case of Linkage Group molecules
            chosen_response = re.findall('.*/chromosome=".+"', assembled_line) or \
                              re.findall('.*/organelle=".+"', assembled_line) or \
                              re.findall('.*/note=".+"', assembled_line)

            # If we have a response to give, no need to continue further
            # If the sixth character is not empty, we have reached the next feature, so no need to continue further
            if chosen_response or line[5:6].strip() != '':
                break
        else:
            line_index += 1

    if not chosen_response:
        return ""

    return str.split(chosen_response[0], '"')[1].strip()


def is_wgs_accession_format(contig_accession):
    """
    Check if a Genbank contig is part of WGS (Whole Genome Shotgun) sequence

    :param contig_accession: Genbank contig accession (ex:  CM003032.1)
    :return: True if the provided contig is in the WGS format
    """
    wgs_prefix = contig_accession[:4]
    wgs_numeric_suffix = contig_accession[4:].replace(".", "")
    return str.isalpha(wgs_prefix) and str.isnumeric(wgs_numeric_suffix)


def get_chromosome_name_for_contig_accession(contig_accession):
    """
    Given a Genbank contig accession, get the corresponding chromosome name

    :param contig_accession: Genbank contig accession (ex:  CM003032.1)
    :return: Chromosome name (ex: 12 when given an accession CM003032.1)
    """

    # Don't bother calling the ENA web service to get the chromosome number if the accession is a WGS accession
    # since the API will proceed to download the entire WGS dataset which can be in hundreds of MBs or even GBs
    # See https://www.ebi.ac.uk/ena/browser/api/text/AABR07050911.1?lineLimit=100&annotationOnly=true for example
    if is_wgs_accession_format(contig_accession):
        return None

    return \
        resolve_contig_accession_to_chromosome_name(contig_accession, 1000) or \
        resolve_contig_accession_to_chromosome_name(contig_accession, 10000) or \
        resolve_contig_accession_to_chromosome_name(contig_accession, 100000)



# Copyright 2022 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from urllib.parse import quote_plus

from ebi_eva_internal_pyutils.config_utils import get_mongo_creds_for_profile, get_accession_pg_creds_for_profile, \
    get_count_service_creds_for_profile, get_properties_from_xml_file, get_variant_load_job_tracker_creds_for_profile


class SpringPropertiesGenerator:
    """
    Class to generate Spring properties for various Spring Batch pipelines.
    These methods can be used to generate complete properties files entirely in Python; alternatively, certain
    properties can be left unfilled and supplied as command-line arguments (e.g. by a NextFlow process).
    """

    def __init__(self, maven_profile, private_settings_file):
        self.maven_profile = maven_profile
        self.private_settings_file = private_settings_file

    @staticmethod
    def _format(*key_value_maps):
        all_params = defaultdict(list)
        for key_value_map in key_value_maps:
            for key in key_value_map:
                if key_value_map[key] is not None:
                    all_params[key.split('.')[0]].append(f'{key}={key_value_map[key]}')
        lines = []
        for key_type in all_params:
            for line in all_params[key_type]:
                lines.append(line)
            lines.append('')

        return '\n'.join(lines)

    @staticmethod
    def _format_str(string, param):
        if param is None:
            return None
        elif not param:
            return ''
        else:
            return string.format(param)

    def _mongo_properties(self):
        mongo_host, mongo_user, mongo_pass = get_mongo_creds_for_profile(
            self.maven_profile, self.private_settings_file)
        username_with_password = (f'{quote_plus(mongo_user)}:{quote_plus(mongo_pass)}@'
                                  if mongo_user is not None and mongo_pass is not None else '')
        return {
            'spring.data.mongodb.uri': f'mongodb://{username_with_password}{mongo_host}/?retryWrites=true&authSource=admin',
        }

    def _variant_load_job_tracker_properties(self):
        variant_url, variant_user, variant_pass = get_variant_load_job_tracker_creds_for_profile(self.maven_profile,
                                                                                                 self.private_settings_file)
        return {
            'job.repository.url': variant_url,
            'job.repository.username': variant_user,
            'job.repository.password': variant_pass,
        }

    def _count_stats_properties(self):
        counts_url, counts_username, counts_password = get_count_service_creds_for_profile(
            self.maven_profile, self.private_settings_file)
        return {
            'eva.count-stats.url': counts_url,
            'eva.count-stats.username': counts_username,
            'eva.count-stats.password': counts_password
        }

    def _common_properties(self, *, read_preference='primary', chunk_size=100, max_pool_size=2):
        """Properties common to all Spring pipelines"""
        props = {
            'spring.datasource.driver-class-name': 'org.postgresql.Driver',
            'spring.datasource.tomcat.max-active': 3,
            'spring.jpa.generate-ddl': 'true',

            'mongodb.read-preference': read_preference,

            'spring.main.web-application-type': 'none',
            'spring.main.allow-bean-definition-overriding': 'true',
            'spring.jpa.properties.hibernate.jdbc.lob.non_contextual_creation': 'true',
            'spring.jpa.properties.hibernate.temp.use_jdbc_metadata_defaults': 'false',
            'spring.jpa.database-platform': 'org.hibernate.dialect.PostgreSQL9Dialect',
            'parameters.chunkSize': chunk_size,
            'spring.datasource.hikari.maximum-pool-size': max_pool_size
        }
        merge = {**self._mongo_properties(), **self._count_stats_properties(), **props}
        return merge

    def _common_accessioning_properties(self, assembly_accession, read_preference, chunk_size):
        pg_url, pg_user, pg_pass = get_accession_pg_creds_for_profile(self.maven_profile, self.private_settings_file)
        accession_db = get_properties_from_xml_file(
            self.maven_profile, self.private_settings_file)['eva.accession.mongo.database']
        props = {
            'spring.datasource.url': pg_url,
            'spring.datasource.username': pg_user,
            'spring.datasource.password': pg_pass,
            'spring.data.mongodb.database': accession_db,
            'parameters.assemblyAccession': assembly_accession,
        }

        merge = {**self._common_properties(read_preference=read_preference, chunk_size=chunk_size), **props}
        return merge

    def _common_accessioning_clustering_properties(self, *, assembly_accession, read_preference, chunk_size):
        """Properties common to accessioning and clustering pipelines."""
        props = {
            'accessioning.submitted.categoryId': 'ss',
            'accessioning.clustered.categoryId': 'rs',
            'accessioning.monotonic.ss.blockSize': 100000,
            'accessioning.monotonic.ss.blockStartValue': 5000000000,
            'accessioning.monotonic.ss.nextBlockInterval': 1000000000,
            'accessioning.monotonic.rs.blockSize': 100000,
            'accessioning.monotonic.rs.blockStartValue': 3000000000,
            'accessioning.monotonic.rs.nextBlockInterval': 1000000000,
            # This value is not used but is required to create beans in Java
            'recovery.cutoff.days': 9999999
        }
        merge = {**self._common_accessioning_properties(assembly_accession, read_preference, chunk_size), **props}
        return merge

    def get_accessioning_properties(self, *, target_assembly=None, fasta=None, assembly_report=None,
                                    project_accession=None, aggregation='BASIC', taxonomy_accession=None,
                                    vcf_file='', output_vcf='', chunk_size=100):
        """Properties for accessioning pipeline."""
        return self._format(
            self._common_accessioning_clustering_properties(assembly_accession=target_assembly,
                                                            read_preference='secondaryPreferred',
                                                            chunk_size=chunk_size),
            {
                'spring.batch.job.names': 'CREATE_SUBSNP_ACCESSION_JOB',
                'parameters.assemblyReportUrl': self._format_str('file:{0}', assembly_report),
                'parameters.contigNaming': 'NO_REPLACEMENT',
                'parameters.fasta': fasta,
                'parameters.forceRestart': 'false',
                'parameters.projectAccession': project_accession,
                'parameters.taxonomyAccession': taxonomy_accession,
                'parameters.vcfAggregation': aggregation,
                'parameters.vcf': vcf_file,
                'parameters.outputVcf': output_vcf
            },
        )

    def get_clustering_properties(self, *, read_preference='primary', job_name=None, source_assembly='',
                                  target_assembly='', rs_report_path='', rs_acc_file='', duplicate_rs_acc_file='',
                                  projects='', project_accession='', vcf=''):
        """Properties common to all clustering pipelines, though not all are always used."""
        return self._format(
            self._common_accessioning_clustering_properties(assembly_accession=target_assembly,
                                                            read_preference=read_preference, chunk_size=100),
            {
                'spring.batch.job.names': job_name,
                'parameters.remappedFrom': source_assembly,
                'parameters.projects': projects,
                'parameters.projectAccession': project_accession,
                'parameters.vcf': vcf,
                'parameters.rsReportPath': rs_report_path,
                'parameters.rsAccFile': rs_acc_file,
                'parameters.duplicateRSAccFile': duplicate_rs_acc_file,
            }
        )

    def get_remapping_extraction_properties(self, *, taxonomy=None, source_assembly=None, fasta=None,
                                            assembly_report=None,
                                            projects='', output_folder=None):
        """Properties for remapping extraction pipeline."""
        return self._format(
            self._common_accessioning_properties(assembly_accession=source_assembly,
                                                 read_preference='secondaryPreferred',
                                                 chunk_size=1000),
            {
                'spring.batch.job.names': 'EXPORT_SUBMITTED_VARIANTS_JOB',
                'parameters.taxonomy': taxonomy,
                'parameters.fasta': fasta,
                'parameters.assemblyReportUrl': self._format_str('file:{0}', assembly_report),
                'parameters.projects': projects,
                'parameters.outputFolder': output_folder
            })

    def get_remapping_ingestion_properties(self, *, source_assembly=None, target_assembly=None, vcf=None, load_to=None,
                                           remapping_version=1.0):
        """Properties for remapping ingestion pipeline."""
        return self._format(
            self._common_accessioning_properties(assembly_accession=target_assembly,
                                                 read_preference='secondaryPreferred',
                                                 chunk_size=1000),
            {
                'spring.batch.job.names': 'INGEST_REMAPPED_VARIANTS_FROM_VCF_JOB',
                'parameters.vcf': vcf,
                'parameters.remappedFrom': source_assembly,
                'parameters.loadTo': load_to,
                'parameters.remappingVersion': remapping_version,
            }
        )

    def get_release_properties(self, *, job_name=None, assembly_accession=None, taxonomy_accession=None, fasta=None,
                               assembly_report=None, contig_naming=None, output_folder=None, accessioned_vcf=None,
                               temp_mongo_db=None):
        common_props = self._common_accessioning_properties(assembly_accession=assembly_accession,
                                                            read_preference='secondaryPreferred', chunk_size=1000)
        # For release in Embassy only
        if temp_mongo_db:
            common_props['spring.data.mongodb.database'] = temp_mongo_db
            common_props['mongodb.read-preference'] = 'primaryPreferred'
            common_props.pop('spring.data.mongodb.host')
            common_props.pop('spring.data.mongodb.port')
            common_props.pop('spring.data.mongodb.username')
            common_props.pop('spring.data.mongodb.password')
        return self._format(
            common_props,
            {
                'spring.batch.job.names': job_name,
                'parameters.taxonomyAccession': taxonomy_accession,
                'parameters.contigNaming': contig_naming,
                'parameters.fasta': fasta,
                'parameters.assemblyReportUrl': self._format_str('file:{0}', assembly_report),
                'parameters.outputFolder': output_folder,
                'parameters.accessionedVcf': '' if accessioned_vcf is None else accessioned_vcf,
                'logging.level.uk.ac.ebi.eva.accession.release': 'INFO'
            })

    def _common_eva_pipeline_properties(self, opencga_path, read_preference='secondaryPreferred'):
        files_collection = get_properties_from_xml_file(
            self.maven_profile, self.private_settings_file)['eva.mongo.collections.files']
        annotation_metadata_collection = get_properties_from_xml_file(
            self.maven_profile, self.private_settings_file)['eva.mongo.collections.annotation-metadata']
        annotation_collection = get_properties_from_xml_file(
            self.maven_profile, self.private_settings_file)['eva.mongo.collections.annotations']
        variants_collection = get_properties_from_xml_file(
            self.maven_profile, self.private_settings_file)['eva.mongo.collections.variants']
        job_tracker_properties = self._variant_load_job_tracker_properties()
        props = {
            'spring.profiles.active': 'production,mongo',
            'spring.profiles.include': 'variant-writer-mongo,variant-annotation-mongo',

            'spring.data.mongodb.authentication-mechanism': 'SCRAM-SHA-1',
            'job.repository.driverClassName': 'org.postgresql.Driver',

            'db.collections.variants.name': variants_collection,
            'db.collections.files.name': files_collection,
            'db.collections.annotation-metadata.name': annotation_metadata_collection,
            'db.collections.annotations.name': annotation_collection,

            'app.opencga.path': opencga_path,
            'config.restartability.allow': 'false',
            'config.db.read-preference': read_preference,

            'logging.level.embl.ebi.variation.eva': 'DEBUG',
            'logging.level.org.opencb.opencga': 'DEBUG',
            'logging.level.org.springframework': 'INFO',
        }

        merge = {**self._common_properties(read_preference=read_preference, chunk_size=100), **props,
                 **job_tracker_properties}
        return merge

    def get_accession_import_properties(self, opencga_path, read_preference='secondaryPreferred'):
        return self._format(self._common_eva_pipeline_properties(opencga_path, read_preference))

    def get_variant_load_properties(self, project_accession, study_name, output_dir, annotation_dir, stats_dir,
                                    vep_cache_path, opencga_path, read_preference='secondaryPreferred'):
        return self._format(
            self._common_eva_pipeline_properties(opencga_path, read_preference),
            {
                'annotation.overwrite': False,
                'app.vep.cache.path': vep_cache_path,
                'app.vep.num-forks': 4,
                'app.vep.timeout': 500,
                'config.chunk.size': 200,

                'input.study.id': project_accession,
                'input.study.name': study_name,
                'input.study.type': 'COLLECTION',

                'output.dir': str(output_dir),
                'output.dir.annotation': str(annotation_dir),
                'output.dir.statistics': str(stats_dir),

                'statistics.skip': False
            },
        )



import os
from distutils.core import setup

from setuptools import find_packages

setup(
    name='ebi_eva_common_pyutils',
    scripts=[os.path.join(os.path.dirname(__file__), 'ebi_eva_internal_pyutils', 'archive_directory.py')],
    packages=find_packages(),
    version='0.7.0',
    license='Apache',
    description='EBI EVA - Common Python Utilities',
    url='https://github.com/EBIVariation/eva-common-pyutils',
    keywords=['EBI', 'EVA', 'PYTHON', 'UTILITIES'],
    install_requires=['requests==2.*', 'lxml>4.9,==4.*', 'pyyaml==6.*', 'cached-property==1.5.*', 'retry>0.9,==0.*',
                      'openpyxl==3.*'],
    extras_require={'eva-internal': ['psycopg2-binary', 'pymongo', 'networkx<=2.5']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3'
    ]
)



import re

import requests
from retry import retry

from ebi_eva_common_pyutils.logger import logging_config as log_cfg


logger = log_cfg.get_logger(__name__)

eutils_url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'
esearch_url = eutils_url + 'esearch.fcgi'
esummary_url = eutils_url + 'esummary.fcgi'
efetch_url = eutils_url + 'efetch.fcgi'
ensembl_url = 'http://rest.ensembl.org/info/assembly'


@retry(tries=3, delay=2, backoff=1.2, jitter=(1, 3))
def get_ncbi_assembly_dicts_from_term(term, api_key=None):
    """Function to return NCBI assembly objects in the form of a list of dictionaries based on a search term."""
    payload = {'db': 'Assembly', 'term': '"{}"'.format(term), 'retmode': 'JSON'}
    if api_key:
        payload['api_key'] = api_key
    req = requests.get(esearch_url, params=payload)
    req.raise_for_status()
    data = req.json()
    assembly_dicts = []
    if data:
        assembly_id_list = data.get('esearchresult').get('idlist')
        payload = {'db': 'Assembly', 'id': ','.join(assembly_id_list), 'retmode': 'JSON'}
        if api_key:
            payload['api_key'] = api_key
        req = requests.get(esummary_url, params=payload)
        req.raise_for_status()
        summary_list = req.json()
        for assembly_id in summary_list.get('result', {}).get('uids', []):
            assembly_dicts.append(summary_list.get('result').get(assembly_id))
    return assembly_dicts


@retry(tries=3, delay=2, backoff=1.2, jitter=(1, 3))
def get_ncbi_taxonomy_dicts_from_term(term, api_key=None):
    """Function to return NCBI taxonomy objects in the form of a list of dictionaries based on a search term."""
    payload = {'db': 'Taxonomy', 'term': '"{}"'.format(term), 'retmode': 'JSON'}
    if api_key:
        payload['api_key'] = api_key
    req = requests.get(esearch_url, params=payload)
    req.raise_for_status()
    data = req.json()
    taxonomy_dicts = []
    if data:
        taxonomy_dicts = get_ncbi_taxonomy_dicts_from_ids(data.get('esearchresult').get('idlist'))
    return taxonomy_dicts


@retry(tries=3, delay=2, backoff=1.2, jitter=(1, 3))
def get_ncbi_taxonomy_dicts_from_ids(taxonomy_ids, api_key=None):
    """Function to return NCBI taxonomy objects in the form of a list of dictionaries
    based on a list of taxonomy ids."""
    taxonomy_dicts = []
    payload = {'db': 'Taxonomy', 'id': ','.join(taxonomy_ids), 'retmode': 'JSON'}
    if api_key:
        payload['api_key'] = api_key
    req = requests.get(esummary_url, params=payload)
    req.raise_for_status()
    summary_list = req.json()
    for taxonomy_id in summary_list.get('result', {}).get('uids', []):
        taxonomy_dicts.append(summary_list.get('result').get(taxonomy_id))
    return taxonomy_dicts


def get_ncbi_assembly_name_from_term(term, api_key=None):
    assembl_dicts = get_ncbi_assembly_dicts_from_term(term, api_key=api_key)
    assembly_names = set([d.get('assemblyname') for d in assembl_dicts])
    if len(assembly_names) > 1:
        # Only keep the one that have the assembly accession as a synonymous and check again
        assembly_names = set([d.get('assemblyname') for d in assembl_dicts
                              if term in d['synonym'].values() or term == d['assemblyaccession']])
    if len(assembly_names) != 1:
        raise ValueError(f'Cannot resolve assembly name for assembly {term} in NCBI. '
                         f'Found {",".join([str(a) for a in assembly_names])}')
    return assembly_names.pop() if assembly_names else None


def retrieve_species_scientific_name_from_tax_id_ncbi(taxid, api_key=None):
    payload = {'db': 'Taxonomy', 'id': taxid}
    if api_key:
        payload['api_key'] = api_key
    r = requests.get(efetch_url, params=payload)
    match = re.search('<Rank>(.+?)</Rank>', r.text, re.MULTILINE)
    rank = None
    if match:
        rank = match.group(1)
    if rank not in ['species', 'subspecies']:
        logger.warning('Taxonomy id %s does not point to a species', taxid)
    match = re.search('<ScientificName>(.+?)</ScientificName>', r.text, re.MULTILINE)
    if match:
        return match.group(1)


def get_species_name_from_ncbi(assembly_acc, api_key=None):
    # We first need to search for the species associated with the assembly
    assembly_dicts = get_ncbi_assembly_dicts_from_term(assembly_acc, api_key=api_key)
    taxids = set([assembly_dict.get('taxid')
        for assembly_dict in assembly_dicts
        if assembly_dict.get('assemblyaccession') == assembly_acc or
           assembly_dict.get('synonym', {}).get('genbank') == assembly_acc])

    # This is a search so could retrieve multiple results
    if len(taxids) != 1:
        raise ValueError(f'Multiple species found for {assembly_acc}. '
                         f'Cannot resolve single species for assembly {assembly_acc} in NCBI.')

    taxonomy_id = taxids.pop()

    scientific_name = retrieve_species_scientific_name_from_tax_id_ncbi(taxonomy_id, api_key=api_key)
    return scientific_name.replace(' ', '_').lower()



# Copyright 2022 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from ebi_eva_common_pyutils.logger import AppLogger
import requests
from retry import retry


class InternalServerError(Exception):
    pass


CONTING_ALIAS_URL = 'https://www.ebi.ac.uk/eva/webservices/contig-alias'


# TODO add the get methods
class ContigAliasClient(AppLogger):
    """
    Python client for interfacing with the contig alias service.
    Authentication is required if using admin endpoints.
    """

    def __init__(self, base_url=None, username=None, password=None, default_page_size=1000):
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = os.environ.get('CONTING_ALIAS_URL') or CONTING_ALIAS_URL
        # Used for get method
        self.default_page_size=default_page_size
        # Only required for admin endpoints
        self.username = username
        self.password = password

    def check_auth(self):
        if self.username is None or self.password is None:
            raise ValueError('Need admin username and password for this method')

    @retry(InternalServerError, tries=3, delay=2, backoff=1.5, jitter=(1, 3))
    def insert_assembly(self, assembly):
        self.check_auth()
        full_url = os.path.join(self.base_url, f'v1/admin/assemblies/{assembly}')

        response = requests.put(full_url, auth=(self.username, self.password))
        if response.status_code == 200:
            self.info(f'Assembly accession {assembly} successfully added to Contig-Alias DB')
        elif response.status_code == 409:
            self.warning(f'Assembly accession {assembly} already exists in Contig-Alias DB. Response: {response.text}')
        elif response.status_code == 500:
            self.error(f'Could not save Assembly accession {assembly} to Contig-Alias DB. Error: {response.text}')
            raise InternalServerError
        else:
            self.error(f'Could not save Assembly accession {assembly} to Contig-Alias DB. Error: {response.text}')
            response.raise_for_status()

    @retry(InternalServerError, tries=3, delay=2, backoff=1.5, jitter=(1, 3))
    def delete_assembly(self, assembly):
        self.check_auth()
        full_url = os.path.join(self.base_url, f'v1/admin/assemblies/{assembly}')

        response = requests.delete(full_url, auth=(self.username, self.password))
        if response.status_code == 200:
            self.info(f'Assembly accession {assembly} successfully deleted from Contig-Alias DB')
        elif response.status_code == 500:
            self.error(f'Assembly accession {assembly} could not be deleted. Response: {response.text}')
            raise InternalServerError
        else:
            self.error(f'Assembly accession {assembly} could not be deleted. Response: {response.text}')

    @retry(tries=3, delay=2, backoff=1.2, jitter=(1, 3))
    def _get_page_for_contig_alias_url(self, sub_url, page=0):
        """queries the contig alias to retrieve the page of the provided url"""
        url = f'{self.base_url}/{sub_url}?page={page}&size={self.default_page_size}'
        response = requests.get(url, headers={'accept': 'application/json'})
        response.raise_for_status()
        response_json = response.json()
        return response_json

    def _depaginate_iter(self, sub_url, entity_to_retrieve):
        """Generator that provides the contigs in the assembly requested."""
        page = 0
        response_json = self._get_page_for_contig_alias_url(sub_url, page=page)
        for entity in response_json.get('_embedded', {}).get(entity_to_retrieve, []):
            yield entity
        while 'next' in response_json['_links']:
            page += 1
            response_json = self._get_page_for_contig_alias_url(sub_url, page=page)
            for entity in response_json.get('_embedded', {}).get(entity_to_retrieve, []):
                yield entity

    def assembly_contig_iter(self, assembly_accession):
        """Generator that provides the contigs in the assembly requested."""
        sub_url = f'v1/assemblies/{assembly_accession}/chromosomes'
        return self._depaginate_iter(sub_url, 'chromosomeEntities')

    def assembly(self, assembly_accession):
        """provides the description of the requested assembly."""
        sub_url = f'v1/assemblies/{assembly_accession}'
        response_json = self._get_page_for_contig_alias_url(sub_url)
        return response_json.get('_embedded', {}).get('assemblyEntities', [])[0]

    def contig_iter(self, insdc_accession):
        sub_url = f'v1/chromosomes/genbank/{insdc_accession}'
        return self._depaginate_iter(sub_url, 'chromosomeEntities')



# Copyright 2020 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http
import requests

from ebi_eva_common_pyutils.assembly import NCBIAssembly
from ebi_eva_common_pyutils.ena_utils import download_xml_from_ena
from ebi_eva_common_pyutils.logger import logging_config as log_cfg

EUTILS_URL = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'
ESEARCH_URL = EUTILS_URL + 'esearch.fcgi'
ESUMMARY_URL = EUTILS_URL + 'esummary.fcgi'
EFETCH_URL = EUTILS_URL + 'efetch.fcgi'


logger = log_cfg.get_logger(__name__)


def is_patch_assembly(assembly_accession: str) -> bool:
    """
    Check if a given assembly is a patch assembly
    Please see: https://www.ncbi.nlm.nih.gov/grc/help/patches/
    """
    xml_root = download_xml_from_ena(f'https://www.ebi.ac.uk/ena/browser/api/xml/{assembly_accession}')
    xml_assembly = xml_root.xpath("//ASSEMBLY_ATTRIBUTE[TAG='count-patches']/VALUE")
    if len(xml_assembly) == 0:
        return False
    return int(xml_assembly[0].text) > 0


def retrieve_genbank_assembly_accessions_from_ncbi(assembly_txt, api_key=None):
    """
    Attempt to find any assembly genebank accession base on a free text search.
    """
    assembly_accessions = set()
    payload = {'db': 'Assembly', 'term': '"{}"'.format(assembly_txt), 'retmode': 'JSON'}
    if api_key:
        payload['api_key'] = api_key
    data = requests.get(ESEARCH_URL, params=payload).json()
    if data and data.get('esearchresult', {}).get('idlist'):
        assembly_id_list = data.get('esearchresult').get('idlist')
        payload = {'db': 'Assembly', 'id': ','.join(assembly_id_list), 'retmode': 'JSON'}
        if api_key:
            payload['api_key'] = api_key
        summary_list = requests.get(ESUMMARY_URL, params=payload).json()
        for assembly_id in summary_list.get('result', {}).get('uids', []):
            assembly_info = summary_list.get('result').get(assembly_id)
            if 'genbank' in assembly_info['synonym']:
                assembly_accessions.add(assembly_info['synonym']['genbank'])
    if len(assembly_accessions) != 1:
        logger.warning('%s Genbank synonyms found for assembly %s ', len(assembly_accessions), assembly_txt)
    return list(assembly_accessions)


def retrieve_genbank_equivalent_for_GCF_accession(assembly_accession, ncbi_api_key=None):
    genbank_synonyms = retrieve_genbank_assembly_accessions_from_ncbi(assembly_accession, api_key=ncbi_api_key)
    if len(genbank_synonyms) != 1:
        raise ValueError('%s Genbank synonyms found for assembly %s ' % (len(genbank_synonyms), assembly_accession))
    return genbank_synonyms.pop()


def resolve_assembly_name_to_GCA_accession(assembly_name):
    ENA_ASSEMBLY_NAME_QUERY_URL = "https://www.ebi.ac.uk/ena/portal/api/search" \
                                  "?result=assembly&query=assembly_name%3D%22{0}%22&format=json".format(assembly_name)
    response = requests.get(ENA_ASSEMBLY_NAME_QUERY_URL)
    if response.status_code == http.HTTPStatus.OK.value:
        response_json = response.json()
        if len(response_json) == 0:
            raise ValueError("Could not resolve assembly name {0} to a GCA accession!".format(assembly_name))
        elif len(response_json) > 1:
            raise ValueError("Assembly name {0} resolved to more than one GCA accession!".format(assembly_name))
        else:
            return response.json()[0]["accession"] + "." + response.json()[0]["version"]
    else:
        raise ValueError("Could not resolve assembly name {0} to a GCA accession!".format(assembly_name))


def get_assembly_report_url(assembly_accession):
    return NCBIAssembly(assembly_accession).assembly_report_url



import re

from openpyxl.reader.excel import load_workbook


def metadata_xlsx_version(metadata_xlsx):
    workbook = load_workbook(metadata_xlsx)
    try:
        instructions_sheet = workbook['PLEASE READ FIRST']
        xlsx_sheet_version_value = instructions_sheet[3][0].value
        match = re.search(r'(\d+\.\d+\.\d+)', '' if xlsx_sheet_version_value is None else xlsx_sheet_version_value)
        xlsx_version = match.group(1) if match else None
    except (KeyError, IndexError):
        xlsx_version = None
    return xlsx_version